{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "# This notebook is associated to the paper \"Counting points on smooth plane quartics\"\n",
    "# by Edgar Costa, David Harvey, and Andrew V. Sutherland\n",
    "# It includes a demonstration implementation of Algorithms 5.8,5.11,5.14\n",
    "# For algorithm 5.16 we give only the simple O(p log p loglog p) implementation\n",
    "# we also implement the uncompressed variation, see remark 5.17.\n",
    "# For the average polynomial time algorithms we implement a simple remainder tree, not a remainder forest (so kappa=0),\n",
    "# We implement:\n",
    "# - the compressed version algorithm 5.22 with remark 5.26 incorporated\n",
    "# - the uncompressed version, see remark 5.24\n",
    "# - optimised Harvey's algorithm with remark 5.29 incorporated.\n",
    "# We only address primes p <= N that do not divide D\n",
    "\n",
    "from sage.all import (\n",
    "    GCD,\n",
    "    GF,\n",
    "    Matrix,\n",
    "    PolynomialRing,\n",
    "    ZZ,\n",
    "    cached_method,\n",
    "    identity_matrix,\n",
    "    inverse_mod,\n",
    "    is_prime,\n",
    "    polygen,\n",
    "    prime_range,\n",
    "    prod,\n",
    "    vector,\n",
    ")\n",
    "\n",
    "\n",
    "def decrease(w, i):\n",
    "    return tuple(c - int(i==j) for j, c in enumerate(w))\n",
    "\n",
    "\n",
    "@cached_method\n",
    "def monomials(d):\n",
    "    return [\n",
    "        (e0, e1, e2)\n",
    "        for e0 in range(d + 1)\n",
    "        for e1 in range(d + 1)\n",
    "        for e2 in range(d + 1)\n",
    "        if e0 + e1 + e2 == d\n",
    "    ]\n",
    "\n",
    "def invfactorialproduct(R, y):\n",
    "    if any(elt not in ZZ or elt < 0 for elt in y):\n",
    "        return 0\n",
    "    y = [elt for elt in y if y != 0]\n",
    "    if not y:\n",
    "        return 1\n",
    "    return 1/prod(prod( R(i) for i in range(1, elt + 1) ) for elt in y)\n",
    "\n",
    "\n",
    "def random_int(bits):\n",
    "    return ZZ.random_element(2 ** (bits + 1) - 1) - 2**bits + 1\n",
    "\n",
    "\n",
    "def random_quartic(bits):\n",
    "    x0, x1, x2 = PolynomialRing(ZZ, \"x\", 3).gens()\n",
    "    return sum(\n",
    "        [\n",
    "            random_int(bits) * x0**e0 * x1**e1 * x2**e2\n",
    "            for (e0, e1, e2) in monomials(4)\n",
    "        ]\n",
    "    )\n",
    "\n",
    "\n",
    "def art(M, m, c=1):\n",
    "    \"\"\"Computes [prod([M[i] for i in range(k+1)])/c^k % m[k] for k in range(len(M))] assuming M[i]*M[i+1] is divisible by c\"\"\"\n",
    "    N = len(M)\n",
    "    if N <= 1:\n",
    "        return [M[0] % m[0]] if N == 1 else []\n",
    "    n = (N + (N % 2)) // 2\n",
    "    M.append(c)\n",
    "    m.append(1)\n",
    "    # pad\n",
    "    C = art(\n",
    "        [(M[2 * k - 1] if k else c) * M[2 * k] / c for k in range(n)],\n",
    "        [m[2 * k] * m[2 * k + 1] for k in range(n)],\n",
    "        c=c,\n",
    "    )\n",
    "    C = sum(\n",
    "        [\n",
    "            [\n",
    "                C[k] % m[2 * k],\n",
    "                C[k] * M[2 * k + 1] * inverse_mod(c, m[2 * k + 1]) % m[2 * k + 1],\n",
    "            ]\n",
    "            for k in range(n)\n",
    "        ],\n",
    "        [],\n",
    "    )[:N]\n",
    "    return C\n",
    "\n",
    "\n",
    "class Quartic:\n",
    "    # F = homogeneous quartic in Z[x0,x1,x2]\n",
    "    def __init__(self, F):\n",
    "        self.F = F\n",
    "        self.d = d = F.degree()\n",
    "        assert d == 4\n",
    "\n",
    "        self.ZZV = PolynomialRing(ZZ, 3, \"V\")\n",
    "        (V0, V1, V2) =  V = self.ZZV.gens()\n",
    "        self.D = [monomials(i) for i in range(2*d)]\n",
    "        self.e = [vector(elt) for elt in reversed(sorted(self.D[1]))]\n",
    "\n",
    "        # for any v with |v| = 4m+6,\n",
    "        # let b(v) = ((F^(n+1))_{v-u})_{u in T2}, ((m + 1) (F^n)_{v-u})_{u in B6}   (vector in Z^16)\n",
    "        # where B6 is a certain 10-element subset of T6.\n",
    "        # note that ell=2d-2=6 in the paper\n",
    "\n",
    "        # compute:\n",
    "        # - the set B6\n",
    "        # - the nonzero integer lambda6,\n",
    "        # - a 16*28 integer matrix PI\n",
    "        # - a 28*16 matrix matrix PSI\n",
    "        # PSI over Z[V0,V1,V2] (with entries degree at most 1) such that\n",
    "        # for any |v| = 4m+6 we have\n",
    "        #   PI a(v) = b(v)\n",
    "        #   PSI(v) b(v) = d (m + 1) lambda6 a(v).\n",
    "        # In particular,\n",
    "        #   PI PSI(v) = d (m + 1) lambda6 I_{16}    (identity matrix)\n",
    "        #   PSI(v) PI a(v) = d (m + 1) lambda_6 a(v)\n",
    "        # (the idea is that PI compresses a length 28 vector into a length 16 vector,\n",
    "        # and PSI(v) performs the reverse operation)\n",
    "\n",
    "        # compute the set B6\n",
    "        T = Matrix(ZZ, 18, 28)\n",
    "        for i in range(3):\n",
    "            for u_index, u in enumerate(self.D[2]):\n",
    "                #u = self.D[2][u_index]\n",
    "                for t in self.D[4]:\n",
    "                    j = self.D[6].index((u[0] + t[0], u[1] + t[1], u[2] + t[2]))\n",
    "                    T[6 * i + u_index, j] = t[i] * F[t]\n",
    "\n",
    "        B6 = T.nonpivots()\n",
    "        J = Matrix(ZZ, 10, 28)\n",
    "        for i, v in enumerate(B6):\n",
    "            J[i, v] = 1\n",
    "        S = T.stack(J);\n",
    "\n",
    "        Sadj = S.adjugate()\n",
    "\n",
    "        Sadj /= GCD(Sadj.list())\n",
    "        lambda6I = Sadj*S\n",
    "        lambda6 = ZZ(lambda6I[0,0])\n",
    "        assert lambda6I == lambda6*identity_matrix(28)\n",
    "\n",
    "\n",
    "        PI = Matrix(ZZ, 16, 28)\n",
    "        for u_index in range(6):\n",
    "            u = self.D[2][u_index]\n",
    "            for t in self.D[4]:\n",
    "                PI[u_index, self.D[6].index((u[0] + t[0], u[1] + t[1], u[2] + t[2]))] = F[\n",
    "                    t\n",
    "                ]\n",
    "        for j in range(10):\n",
    "            PI[j + 6, B6[j]] = 1\n",
    "        self.PI = PI\n",
    "\n",
    "        m = -2 # taking advantage that m = p - 2 = -2 mod p\n",
    "        PSI = Matrix(self.ZZV, 28, 16)\n",
    "        V = self.ZZV.gens()\n",
    "        for u_index, u in enumerate(self.D[6]):\n",
    "            # how to write x^u\n",
    "            row = Sadj[u_index]\n",
    "            # the term in D_2\n",
    "            for t_index, t in enumerate(self.D[2]):\n",
    "                PSI[u_index, t_index] = sum((V[i] - t[i]) * row[6*i + t_index] for i in range(3))\n",
    "            # the term in B_6\n",
    "            for b_index, b in enumerate(B6):\n",
    "                PSI[u_index, 6 + b_index] = (m + 1) * row[6*3 + b_index]\n",
    "        self.PSI = PSI\n",
    "        # self.PSIeval = PSI(0, 2*p-1, 2*p - 1) mod p\n",
    "        self.PSIeval = PSI.substitute({V0:0, V1: -1, V2:  - 1}).change_ring(ZZ)\n",
    "        self.c = lambda6  # this is the c in Remark 5.15\n",
    "        self.theta0 = self.phi_generic(0, 1)[0]\n",
    "        self.theta1 = self.phi_generic(1, 0)[0]\n",
    "        self.denominator = 2 * lambda6 * self.theta0 * self.theta1\n",
    "\n",
    "    @cached_method\n",
    "    def phi_generic(self, i, j):\n",
    "        d = self.d\n",
    "        ell = 2*d-2\n",
    "        A = Matrix(2*d, 2*d)\n",
    "        Mconst, Mi, Mk = [Matrix(2*d, (3*d - 1)*d//2) for _ in range(3)]\n",
    "        k = (-i - j) % 3\n",
    "        m = -2 # taking advantage that m = p - 2 = -2 mod p\n",
    "\n",
    "        U = sorted([tuple(a * self.e[k] - ((a + 1) - 4) * self.e[j]) for a in range(d)])\n",
    "        S = sorted([tuple(c * self.e[k] - (c + 1 - 8) * self.e[j]) for c in range(2*d)])\n",
    "        # T = list({tuple(vector(u) - vector(t)) for u in U for t in self.D[4]}.difference(S))\n",
    "        T = sorted({tuple(vector(u) + vector(t)) for u in U for t in self.D[d] if t[i] != 0})\n",
    "        assert len(T) == (3*d - 1)*d//2\n",
    "\n",
    "        for a, u in enumerate(U):\n",
    "            for t, ft in self.F.dict().items():\n",
    "                uplust = tuple(vector(u) + vector(t))\n",
    "                if t[i] == 0:\n",
    "                    Gcoordinate = S.index(uplust)\n",
    "                    A[2 * a, Gcoordinate] = ft\n",
    "                    A[2 * a + 1, Gcoordinate] = t[k] * ft\n",
    "                else:\n",
    "                    Gcoordinate = T.index(uplust)\n",
    "                    # M[2*l][Gcoordinate] = ((m + 1)*t[i] - w[i])*f[t]\n",
    "                    #                     = ((m + 1)*t[i] - (v[i] + 1))*f[t]\n",
    "                    Mconst[2 * a, Gcoordinate] = ((m + 1) * t[i] - 1) * ft\n",
    "                    Mi[2 * a, Gcoordinate] = -ft\n",
    "                    # Mk[2*l][Gcoordinate] = 0\n",
    "\n",
    "                    # M[2*l + 1][Gcoordinate]  =\n",
    "                    # = (w[k]*t[i] - w[i]*t[k])*f[t]\n",
    "                    # = ((v[k] - u[k])*t[i] - (v[i] + 1)*t[k])*f[t]\n",
    "                    Mconst[2 * a + 1, Gcoordinate] = (-u[k] * t[i] -t[k]) * ft\n",
    "                    Mi[2 * a + 1, Gcoordinate] = -t[k] * ft\n",
    "                    Mk[2 * a + 1, Gcoordinate] = t[i] * ft\n",
    "                    # Bdegree[2*l+1] = 0\n",
    "\n",
    "        thetai = A.determinant()\n",
    "        adjA = A.adjugate()\n",
    "\n",
    "        Bconst, Bi, Bk = [adjA * elt for elt in [Mconst, Mi, Mk]]\n",
    "\n",
    "        PHIconst, PHIi, PHIk = [Matrix(len(self.D[ell + 1]), len(self.D[ell])) for _ in range(3)]\n",
    "        for s_coordinate, s in enumerate(self.D[ell + 1]):\n",
    "            if s[i] > 0: # i.e., s - ei in D(v, 6)\n",
    "                si = decrease(s, i)\n",
    "                si_coordinate = self.D[ell].index(si)\n",
    "                PHIconst[s_coordinate, si_coordinate] = thetai\n",
    "                PHIi[s_coordinate, si_coordinate] = thetai\n",
    "            if s[i] == 0:\n",
    "                assert s in S\n",
    "                B_row = S.index(s)\n",
    "                for j, t in enumerate(T):\n",
    "                    ti = decrease(t, i)\n",
    "                    ti_coordinate = self.D[ell].index(ti)\n",
    "                    PHIconst[s_coordinate, ti_coordinate] = Bconst[B_row, j]\n",
    "                    PHIi[s_coordinate, ti_coordinate] = Bi[B_row, j]\n",
    "                    PHIk[s_coordinate, ti_coordinate] = Bk[B_row, j]\n",
    "\n",
    "        return thetai, PHIconst, PHIi, PHIk\n",
    "\n",
    "    @cached_method\n",
    "    def tau_generic(self, i, j):\n",
    "        thetai, PHIconst, PHIi, PHIk = self.phi_generic(i,j)\n",
    "        projT = Matrix([[ int(t==decrease(w,j)) for w in self.D[7]] for t in self.D[6]])\n",
    "        tauconst, taui, tauk = [projT*elt for elt in [PHIconst, PHIi, PHIk]]\n",
    "        return thetai, tauconst, taui, tauk\n",
    "\n",
    "    def tau(self, v, i, j):\n",
    "        k = (-i - j) % 3\n",
    "        thetai, tauconst, taui, tauk = self.tau_generic(i, j)\n",
    "        return thetai, tauconst + taui*v[i] + tauk*v[k], taui\n",
    "\n",
    "    def T(self, v, i, j):\n",
    "        thetai, tauconst, taui, tauk = self.tau_generic(i, j)\n",
    "        V = self.ZZV.gens()\n",
    "        k = (-i - j) % 3\n",
    "        tau = tauconst + taui*V[i] + tauk*V[k]\n",
    "        calT = self.PI.change_ring(self.ZZV)*tau*self.PSI\n",
    "        x = polygen(ZZ)\n",
    "        v = list(v)\n",
    "        v[i] += x\n",
    "        v[j] -= x\n",
    "        calTv = calT(*v)\n",
    "        T = [Matrix(16,16) for _ in range(3)]\n",
    "        for r in range(16):\n",
    "            for c in range(16):\n",
    "                for d in range(3):\n",
    "                    T[d][r,c] = calTv[r,c][d]\n",
    "        return T\n",
    "\n",
    "    @cached_method\n",
    "    def Tk_gen(self):\n",
    "        return self.T((0,-1,-1), 0, 1)\n",
    "\n",
    "    def Tk(self, k):\n",
    "        T0, T1, T2 = self.Tk_gen()\n",
    "        # we realign the k variable so that\n",
    "        # newk = p - 2 - k, so that Tk(0) ... Tk(p-2)\n",
    "        # matches with T_{w + (p-2)(e0 -e1), 0, 1} ... T_{w, 0, 1}\n",
    "        newk = -2 -k\n",
    "        return T0 + newk*T1 + newk**2*T2\n",
    "\n",
    "    def tauk(self, k):\n",
    "        _, T0, T1 = self.tau((0,-1,-1), 0, 1)\n",
    "        # we realign the k variable so that\n",
    "        # newk = p - 2 - k, so that tauk(0) ... tauk(p-2)\n",
    "        # matches with tau_{w + (p-2)(e0 -e1), 0, 1} ... tau_{w, 0, 1}\n",
    "        newk = -2 -k\n",
    "        return T0 + newk*T1\n",
    "\n",
    "    @cached_method\n",
    "    def equation_matrix(self, v):\n",
    "        ell = 2*self.d-2\n",
    "        m = -2\n",
    "        M = Matrix(2*len(self.D[ell - self.d]), len(self.D[ell]))\n",
    "        for s_coordinate, s in enumerate(self.D[ell - self.d]):\n",
    "            for t, ft in self.F.dict().items():\n",
    "                spt = tuple(vector(s) + vector(t))\n",
    "                j = self.D[ell].index(spt)\n",
    "                for i in range(2):\n",
    "                    M[2*s_coordinate + i, j] = ((v[i] - s[i]) - (m + 1) * t[i])*ft\n",
    "        return M\n",
    "\n",
    "    @cached_method\n",
    "    def fpower_naive(self, p):\n",
    "        return self.F.change_ring(GF(p))**(p-2)\n",
    "\n",
    "    def naive_f_at_v(self, p, v):\n",
    "        fm = self.fpower_naive(p)\n",
    "        ell = sum(v) - self.d*(p-2)\n",
    "        coordinates = [tuple(vector(v) - vector(elt)) for elt in self.D[ell]]\n",
    "        return vector([fm[elt] if all(m >= 0 for m in elt) else 0  for elt in coordinates])\n",
    "\n",
    "    def cm_naive(self, p):\n",
    "        \"\"\"Computes the Cartier-Manin matrix diretly from the definition\"\"\"\n",
    "        Fpow = self.F.change_ring(GF(p)) ** (p - 1)\n",
    "        # Apply equation (2.6) in the paper\n",
    "        return Matrix(\n",
    "            [\n",
    "                [Fpow[p-1, p-1, 2*p-2], Fpow[2*p-1, p-1, p-2], Fpow[p-1, 2*p-1, p-2]],\n",
    "                [Fpow[p-2, p-1, 2*p-1], Fpow[2*p-2, p-1, p-1], Fpow[p-2, 2*p-1, p-1]],\n",
    "                [Fpow[p-1, p-2, 2*p-1], Fpow[2*p-1, p-2, p-1], Fpow[p-1, 2*p-2, p-1]]\n",
    "            ]\n",
    "        )\n",
    "\n",
    "    # This is Algorithm 5.8 in the paper with steps (1) and (2) interleaved\n",
    "    def cm_final(self, p, Cp):\n",
    "        \"\"\"Given Cp for a prime p not dividing D, computes the Cartier-Manin matrix Cp\"\"\"\n",
    "        K = GF(p)\n",
    "        A = Matrix(K, 3, 3)\n",
    "        R = PolynomialRing(K, \"x\")\n",
    "        F = self.F\n",
    "\n",
    "        assert Cp.ncols() == Cp.nrows() and Cp.ncols() in [16, 28]\n",
    "\n",
    "        compressed = Cp.ncols() == 16\n",
    "\n",
    "        # (a) compute 1st column of A\n",
    "\n",
    "        # step 1a\n",
    "        # compute a1 = a(w_1) = a(0, 2p-1, 2p-1)\n",
    "        g = R([F[0, 4, 0], F[0, 3, 1], F[0, 2, 2], F[0, 1, 3], F[0, 0, 4]]) ** 2\n",
    "        # compute starting vector F^(p-2)|D(w1,6) via Section 4.2 using g=F(0,1,x)^2\n",
    "        # using coefficients of series (J/J0)^(-1) = 1 + G1*x + G2*x^2 + ...\n",
    "        Q = Matrix(K, 8, 8)\n",
    "        for j in range(8):\n",
    "            Q[0, j] = -g[j + 1] / g[0]\n",
    "        for j in range(7):\n",
    "            Q[j + 1, j] = 1\n",
    "        Qpow = Q ** (p - 1)\n",
    "        Qpow2 = Qpow * Qpow * Q  # M^(2p-1)\n",
    "        # deduce (F^(p-2))_{0, 2p-1-(6-j), 2p-1-j} for 0 <= j <= 6\n",
    "        a1 = vector(K, 28)\n",
    "        for j in range(7):\n",
    "            u = (0, 6 - j, j)\n",
    "            a1[self.D[6].index(u)] = (\n",
    "                Qpow[j, 0] * F[0, 3, 1] / F[0, 4, 0] ** 2 + Qpow2[j, 0] / F[0, 4, 0]\n",
    "            )\n",
    "\n",
    "        # step 2a\n",
    "        # temp = b(p-1, p, 2p-1)\n",
    "        if compressed:\n",
    "            # compress it by using PI\n",
    "            temp = Cp * self.PI * a1\n",
    "        if not compressed:\n",
    "            # go from F^(p-2) to F^(p-1) by applying PI\n",
    "            temp = self.PI * Cp * a1\n",
    "        # first col of CM matrix\n",
    "        A[0, 0] = temp[self.D[2].index((0, 1, 1))]\n",
    "        A[1, 0] = temp[self.D[2].index((1, 1, 0))]\n",
    "        A[2, 0] = temp[self.D[2].index((0, 2, 0))]\n",
    "\n",
    "        # (b) compute 2nd column of A\n",
    "\n",
    "        # step 1b\n",
    "        # compute a2 = a(w_2) = a((3p-1, 0, p-1))\n",
    "        # compute starting vector F^(p-2)|D(w2,6) via Section 4.2 using g=F(1,0,x)^2\n",
    "        g = R([F[4, 0, 0], F[3, 0, 1], F[2, 0, 2], F[1, 0, 3], F[0, 0, 4]]) ** 2\n",
    "        Q = Matrix(K, 8, 8)\n",
    "        for j in range(8):\n",
    "            Q[0, j] = -g[j + 1] / g[0]\n",
    "        for j in range(7):\n",
    "            Q[j + 1, j] = 1\n",
    "        Qpow_2 = Q ** (p - 1)\n",
    "        # deduce (F^(p-2))_{3p-1-(6-j), 0, p-1-j} for 0 <= j <= 6\n",
    "        a2 = vector(K, 28)\n",
    "        for j in range(7):\n",
    "            u = (6 - j, 0, j)\n",
    "            a2[self.D[6].index(u)] = Qpow_2[j, 0] / F[4, 0, 0]\n",
    "\n",
    "        # step 2b\n",
    "        # temp = b(v_2) = b((2p, p-1, p-1))\n",
    "        # (NOTE: in this next line, we don't need the matrix inverse; just the solution to the system is enough)\n",
    "        if compressed:\n",
    "            # compress it by using PI\n",
    "            temp = Cp.solve_right(self.PI * a2)\n",
    "        if not compressed:\n",
    "            # go from F^(p-2) to F^(p-1) by applying PI\n",
    "            # note: Cp is not invertible, but a2 is in its row space\n",
    "            # so if we add the equations that temp must also solve, we get a unique solution\n",
    "            M = Cp.stack(self.equation_matrix((0, -1, -1)))\n",
    "            assert M.rank() == 28\n",
    "            temp = self.PI*M.solve_right(vector(GF(p), a2.list() + [0]*12))\n",
    "        # second col of CM matrix\n",
    "        A[0, 1] = temp[self.D[2].index((1, 0, 1))]\n",
    "        A[1, 1] = temp[self.D[2].index((2, 0, 0))]\n",
    "        A[2, 1] = temp[self.D[2].index((1, 1, 0))]\n",
    "\n",
    "        # (c) compute 3rd column of A\n",
    "\n",
    "        # step 1c\n",
    "        # compute a(w_3) = a((0, 3p-1, p-1))\n",
    "        # we already computed starting vector F^(p-2)|D(w3,6) via Section 4.2 using g=F(0,1,x)^2 in step 1a\n",
    "        # deduce (F^(p-2))_{0, 3p-1-(6-j), p-1-j} for 0 <= j <= 6\n",
    "        a3 = vector(K, 28)\n",
    "        for j in range(7):\n",
    "            u = (0, 6 - j, j)\n",
    "            a3[self.D[6].index(u)] = Qpow[j, 0] / F[0, 4, 0]\n",
    "\n",
    "        # step 2c\n",
    "        # temp = b(v_1) = b(p-1, 2p, p-1) = (p-1, 2p, p-1)\n",
    "        if compressed:\n",
    "            # compress it by using PI\n",
    "            temp = Cp * self.PI * a3\n",
    "        if not compressed:\n",
    "            # go from F^(p-2) to F^(p-1) by applying PI\n",
    "            temp = self.PI * Cp * a3\n",
    "        # third col of CM matrix\n",
    "        A[0, 2] = temp[self.D[2].index((0, 1, 1))]\n",
    "        A[1, 2] = temp[self.D[2].index((1, 1, 0))]\n",
    "        A[2, 2] = temp[self.D[2].index((0, 2, 0))]\n",
    "\n",
    "        return A\n",
    "\n",
    "    def good_primes(self, N):\n",
    "        return [p for p in prime_range(N + 1) if self.denominator % p != 0]\n",
    "\n",
    "    def mn(self, n):\n",
    "        p = n + 2\n",
    "        return p if is_prime(p) and self.denominator % p != 0 else 1\n",
    "\n",
    "    def Cp(self, p, compressed=True):\n",
    "        r\"\"\"\n",
    "        return:\n",
    "        let v = (0,-1,-1) = (0, 2p-1, 2p-1) mod p = (0, 3p-1, p-1) mod p\n",
    "        if compressed:\n",
    "            C[p] = - T_{v + (p-2)(e0 - e1), 0, 1} ... T_{v, 0, 1}\n",
    "        else:\n",
    "            C[p] = - tau_{v + (p-2)(e0 - e1), 0, 1} ... tau_{v, 0, 1}\n",
    "        i.e., the transition matrix from b(0, 2p-1, 2p-1) to b(p-1, p, 2p-1) modulo p\n",
    "        (the -1 comes from the denominator (p-1)! mod p)\n",
    "        \"\"\"\n",
    "        Mn = self.Tk if compressed else self.tauk\n",
    "        return -prod([Mn(n).change_ring(GF(p)) for n in range(p - 1)])\n",
    "\n",
    "    def cartier_manin_matrices(self, N, algorithm=\"CHS\", compressed=True):\n",
    "        \"\"\"returns dictionary p ->  Cartier Manin matrix at p, for p < N not dividing D\"\"\"\n",
    "        if algorithm == \"naive\":\n",
    "            # This is the O(p log p loglog p) implementation of Algorithm 5.11\n",
    "            return {p: self.cm_final(p, self.Cp(p)) for p in self.good_primes(N - 1)}\n",
    "        elif algorithm == \"CHS\":\n",
    "            # This is a remainder tree implementation of Algorithm 5.14 that uses Remark 5.16\n",
    "            if compressed:\n",
    "                M = [self.Tk(n) for n in range(N)]\n",
    "            else:\n",
    "                M = [self.tauk(n) for n in range(N)]\n",
    "            m = [self.mn(n) for n in range(N - 1)]\n",
    "            c = self.c if compressed else 1\n",
    "            C = art(M, m, c=c)\n",
    "            return {\n",
    "                p: self.cm_final(p, -C[p - 2].change_ring(GF(p)) / c)\n",
    "                for p in self.good_primes(N)\n",
    "            }\n",
    "        elif algorithm == \"Harvey\":\n",
    "            d = self.d\n",
    "            h = 3*d-2\n",
    "            z = vector([0,0,h-d])\n",
    "            Bh = [vector(elt) for elt in monomials(h)]\n",
    "            # interior of Newton polygon\n",
    "            Bds = [vector([1,1,2]), vector([2,1,1]), vector([1,2,1])]\n",
    "            V0 = Matrix(len(Bds), len(Bh))\n",
    "            # pick the 3 columns we  want\n",
    "            for i, elt in enumerate(Bds):\n",
    "                V0[i, Bh.index(z + elt)] = 1\n",
    "\n",
    "\n",
    "            M = [self.Qk(n) for n in range(N)]\n",
    "            M[0] = V0*M[0]\n",
    "            m = [self.mn(n) for n in range(N - 1)]\n",
    "            Q = art(M, m, c=1)\n",
    "            return {\n",
    "                p: self.cm_harvey(p, Q[p - 2].change_ring(GF(p)))\n",
    "                for p in self.good_primes(N)\n",
    "            }\n",
    "        else:\n",
    "            raise ValueError(\"not a valid algorithm option\")\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "    @cached_method\n",
    "    def Qmatrix(self):\n",
    "        d = self.d\n",
    "        h = 3*d-2\n",
    "        Bh = [vector(elt) for elt in monomials(h)]\n",
    "        n = len(Bh)\n",
    "        Q0 = Matrix(n, n)\n",
    "        Q1 = Matrix(n, n)\n",
    "        o = [vector(elt) for elt in [(d,0,0), (0,d,0), (0,0,d)]];\n",
    "        for t, bt in enumerate(Bh):\n",
    "            i = min([i for i in range(3) if bt[i] >= d])\n",
    "            for u, bu in enumerate(Bh):\n",
    "                a = vector(bu) - vector(bt) + o[i]\n",
    "                if any(elt < 0 for elt in a):\n",
    "                    continue\n",
    "                Q0[t, u] = self.F[a]*((h if i == 2 else d) - bt[i] - a[i])\n",
    "                Q1[t, u] = - self.F[a]*a[i]\n",
    "        return Q0, Q1\n",
    "\n",
    "\n",
    "    def Qk(self, k):\n",
    "        # we realign the k variable so that\n",
    "        # newk = p - 2 - k, so that Qk(0) ... Qk(p-2)\n",
    "        # matches with Q(p-2) ... Q(0)\n",
    "        newk = -2 -k\n",
    "        Q0, Q1 = self.Qmatrix()\n",
    "        return Q0 + newk*Q1\n",
    "\n",
    "    def cm_harvey(self, p, Qp):\n",
    "        d = self.d\n",
    "        h = 3*d-2\n",
    "        z = vector([0,0,h-d])\n",
    "        Bh = [vector(elt) for elt in monomials(h)]\n",
    "        # interior of Newton polygon\n",
    "        Bds = [vector([1,1,2]), vector([2,1,1]), vector([1,2,1])]\n",
    "        V = Matrix(len(Bh), len(Bds))\n",
    "        for i, bi in enumerate(Bh):\n",
    "            for j, bj in enumerate(Bds):\n",
    "                V[i, j] = invfactorialproduct(GF(p), (p*bj + z -bi)/d)\n",
    "        return Qp*V\n",
    "\n",
    "\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 682 ms, sys: 48 µs, total: 682 ms\n",
      "Wall time: 696 ms\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 481 ms, sys: 0 ns, total: 481 ms\n",
      "Wall time: 504 ms\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 1.49 s, sys: 4.99 ms, total: 1.49 s\n",
      "Wall time: 1.52 s\n",
      "True True\n",
      "True\n"
     ]
    }
   ],
   "source": [
    "#F = random_quartic(10)\n",
    "R.<x0,x1,x2> = ZZ[]\n",
    "F = -60*x0^4 + 982*x0^3*x1 - 1003*x0^2*x1^2 + 918*x0*x1^3 - 265*x1^4 - 889*x0^3*x2 - 995*x0^2*x1*x2 + 116*x0*x1^2*x2 + 457*x1^3*x2 - 938*x0^2*x2^2 + 522*x0*x1*x2^2 - 912*x1^2*x2^2 + 471*x0*x2^3 + 828*x1*x2^3 - 760*x2^4;\n",
    "X = Quartic(F)\n",
    "expected =  {7: 6, 13: 10, 17: 7, 23: 3, 29: 23, 31: 4, 37: 32, 41: 5, 43: 0, 47: 5, 59: 53, 61: 57, 67: 4, 71: 12, 73: 66, 79: 11, 89: 78, 97: 2, 101: 9, 103: 99, 107: 1, 109: 32, 113: 24, 127: 7, 131: 121, 137: 0, 139: 13, 149: 23, 151: 14, 163: 16, 167: 7, 173: 9, 179: 10, 181: 0, 191: 6, 193: 6, 197: 17, 199: 184}\n",
    "%time CM_naive = X.cartier_manin_matrices(200, algorithm=\"naive\")\n",
    "%time CM_CHS =  X.cartier_manin_matrices(200, algorithm=\"CHS\")\n",
    "%time CM_H = X.cartier_manin_matrices(200, algorithm=\"Harvey\")\n",
    "\n",
    "print(CM_naive == CM_CHS, CM_naive == CM_H)\n",
    "print({p: A.trace() for p, A in CM_CHS.items()} == {p:expected[p] for p in expected if p <= 200})"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 3.57 s, sys: 26 ms, total: 3.59 s\n",
      "Wall time: 3.64 s\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 7.76 s, sys: 34.5 ms, total: 7.79 s\n",
      "Wall time: 8.17 s\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 14.7 s, sys: 17.1 ms, total: 14.7 s\n",
      "Wall time: 15.2 s\n",
      "True True\n"
     ]
    }
   ],
   "source": [
    "%time CM_compressed = X.cartier_manin_matrices(2^10)\n",
    "%time CM_uncompressed = X.cartier_manin_matrices(2^10, compressed=False)\n",
    "%time CM_H = X.cartier_manin_matrices(2^10, algorithm=\"naive\")\n",
    "print(CM_compressed == CM_uncompressed, CM_compressed == CM_H)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 7.36 s, sys: 54.1 ms, total: 7.41 s\n",
      "Wall time: 7.52 s\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 19.6 s, sys: 113 ms, total: 19.7 s\n",
      "Wall time: 20.6 s\n"
     ]
    }
   ],
   "source": [
    "%time foo = X.cartier_manin_matrices(2^11)\n",
    "%time foo = X.cartier_manin_matrices(2^11, compressed=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 16.9 s, sys: 78.1 ms, total: 17 s\n",
      "Wall time: 17.6 s\n"
     ]
    }
   ],
   "source": [
    "%time foo = X.cartier_manin_matrices(2^12)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 38.8 s, sys: 445 ms, total: 39.2 s\n",
      "Wall time: 40.9 s\n"
     ]
    }
   ],
   "source": [
    "%time foo = X.cartier_manin_matrices(2^13)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 1min 26s, sys: 841 ms, total: 1min 27s\n",
      "Wall time: 1min 30s\n"
     ]
    }
   ],
   "source": [
    "%time foo = X.cartier_manin_matrices(2^14)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "CPU times: user 4min 18s, sys: 1.41 s, total: 4min 20s\n",
      "Wall time: 4min 30s\n"
     ]
    }
   ],
   "source": [
    "%time foo = X.cartier_manin_matrices(2^14,  compressed=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 0,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "SageMath 9.4",
   "language": "sagemath",
   "metadata": {
    "cocalc": {
     "description": "Open-source mathematical software system",
     "priority": 1,
     "url": "https://www.sagemath.org/"
    }
   },
   "name": "sage-9.4",
   "resource_dir": "/ext/jupyter/kernels/sage-9.4"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}